/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License (GPL3), or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Library General Public License 3 for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License 3 along with this program; if not, see /usr/share/common-licenses/GPL file
 * or write to Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/
#include <math.h>
#include "gst-pipeline.h"
#include "audio-sources.h"

/*
 Tested formats from gconf-editor,  system -> gstreamer -> 0.10 -> audio -> profiles:
 MP2: audio/x-raw,rate=44100,channels=2 ! twolame name=enc mode=0 bitrate=192 ! id3v2mux
 MP3: audio/x-raw,rate=44100,channels=2 ! lamemp3enc name=enc target=0 quality=6 ! xingmux ! id3v2mux
 FLAC: audio/x-raw,rate=44100,channels=2 ! flacenc name=enc
 WAV: audio/x-raw,rate=22050,channels=1 ! wavenc name=enc

 AAC: audio/x-raw,rate=44100,channels=2 ! faac profile=2 ! ffmux_mp4
 OGG: audio/x-raw,rate=44100,channels=2 ! vorbisenc name=enc quality=0.5 ! oggmux
 SPX: audio/x-raw,rate=32000,channels=1 ! speexenc name=enc ! oggmux
*/

//static GstElement *pipeline_create_TEST(PipelineParms *parms, gchar **err_msg);
static GstElement *pipeline_create_simple(PipelineParms *parms, gchar **err_msg);
static GstElement *pipeline_create_complex(PipelineParms *parms, gchar **err_msg);

//static GstElement *pipeline_create_simple_VAD(PipelineParms *parms, gchar **err_msg);
static GstElement *pipeline_create_complex_VAD(PipelineParms *parms, gchar **err_msg);

gchar *pipeline_get_debug_str(GstElement *pipeline);

void pipeline_free_parms(PipelineParms *parms) {
    if (!parms) return;
    g_free(parms->source);
    str_list_free(parms->dev_list);
    g_free(parms->profile_str);
    g_free(parms->file_ext);
    g_free(parms->filename);

    g_free(parms);
}

static GstElement *create_element(const gchar *elem, const gchar *name) {
    GstElement *e = gst_element_factory_make(elem, name);
    if (!GST_IS_ELEMENT(e)) {
        LOG_ERROR("Cannot create element \"%s\" (%s).\n", elem, name);
        return NULL;
    }
    return e;
}

GstElement *pipeline_create(PipelineParms *parms, gchar **err_msg) {
    if (!parms) return NULL;

    // Create a GStreamer pipeline for audio recording
    GstElement *pipeline = NULL;

    // Wash the device list. User may have disconnected microphones and webcams.
    // Invalid devices will crash the GStreamer pipeline.
    // Remove invalid devices.
    GList *new_list = audio_sources_wash_device_list(parms->dev_list);

    if (g_list_length(new_list) < 1) {
        new_list = g_list_append(new_list, NULL);
    }

    // Zero or one device?
    if (g_list_length(new_list) < 2) {

        // Create a simple pipeline that can record from max 1 device
        pipeline = pipeline_create_simple(parms, err_msg);

    } else {

        // Create a complex pipeline that can record from 2 or more devices
        pipeline = pipeline_create_complex(parms, err_msg);
    }

    // Free new_list
    str_list_free(new_list);
    new_list = NULL;

    return pipeline;
}

static GstElement *pipeline_create_simple(PipelineParms *parms, gchar **err_msg) {
    // Create a simple pipeline that can record from one (1) device.
    // 
    // Typical pipeline:
    // $ gst-launch-0.10 pulsesrc device=alsa_output.pci-0000_04_02.0.analog-stereo.monitor
    //        ! queue
    //        ! level name=level
    //        ! audioconvert ! vorbisenc ! oggmux ! filesink location=test1.ogg
    //
    // Available devices (get listing from pulseaudio):
    // $ pactl list | grep -A2 'Source #' | grep 'Name: ' | cut -d" " -f2

    GstElement *pipeline = gst_pipeline_new("Audio-Recorder");

    // Source
    const gchar *source_name = (parms->source ? parms->source : "pulsesrc");
    GstElement *source = create_element(source_name, NULL);

    // Set device
    const gchar *device = g_list_nth_data(parms->dev_list, 0);
    if (device) {
        g_object_set(G_OBJECT(source), "device", device, NULL);
    }

    GstElement *level = create_element("level", "level");

    //GstElement *level = create_element("level", "level");

    // Create a GstCapsfilter + all encoder elements from the GNOME's profile_str.
    // See: gconf-editor,  system -> gstreamer -> 0.10 -> audio -> profiles.
    gchar *str = g_strdup_printf("capsfilter caps=%s", parms->profile_str);

    GError *error = NULL;
    GstElement *bin = gst_parse_bin_from_description(str, TRUE, &error);
    if (error) {
        // Set err_msg
        gchar *tmp = g_strdup_printf("%s. (%s)", error->message, str);
        *err_msg = g_strdup_printf(_("Cannot create audio pipeline. %s.\n"), tmp);
        g_free(tmp);
        g_error_free(error);
        g_free(str);
        goto LBL_1;
    }

    g_free(str);

    GstElement *resample = create_element("audioresample", NULL);
    GstElement *convert = create_element("audioconvert", NULL);

    // Filesink. Caller must set its "location" property.
    GstElement *filesink = create_element("filesink", "filesink");

    gst_bin_add_many(GST_BIN(pipeline), source, level, resample, convert, bin, filesink, NULL);

    // Link
    if (!gst_element_link_many(source, level, resample, convert, bin, filesink, NULL)) {
        *err_msg = g_strdup_printf(_("Cannot create audio pipeline. %s.\n"), "Cannot link.");
        goto LBL_1;
    }

    // Ok
    return pipeline;

LBL_1:
    // Got an error
    return NULL;
}

static GstElement *pipeline_create_complex(PipelineParms *parms, gchar **err_msg) {
    // Create a complex pipeline using the GstAdder element.
    // Ref: http://gstreamer.freedesktop.org/data/doc/gstreamer/head/gst-plugins-base-plugins/html/gst-plugins-base-plugins-adder.html
    // This can record from 2, 3 or more devices.

    // Typical pipeline:
    // $ gst-launch-0.10 adder name=mix
    //        ! level name=level
    //        ! audioconvert ! vorbisenc ! oggmux ! filesink location=test1.ogg
    //        { pulsesrc device=alsa_output.pci-0000_04_02.0.analog-stereo.monitor ! queue ! mix. }
    //        { pulsesrc device=alsa_input.pci_8086_24c5_alsa_WebCam ! queue ! mix.
    //
    // Available devices (get listing from pulseaudio):
    // $ pactl list | grep -A2 'Source #' | grep 'Name: ' | cut -d" " -f2

    GstElement *pipeline = gst_pipeline_new("Audio-Recorder");

    // Mix audio from several sources
    GstElement *adder = create_element("adder", "mixer");

    // Level (dB) data
    GstElement *level = create_element("level", "level");

    // Create a GstCapsfilter + all encoder elements from the GNOME's profile_str.
    // See: gconf-editor,  system -> gstreamer -> 0.10 -> audio -> profiles.
    gchar *str = g_strdup_printf("capsfilter caps=%s", parms->profile_str);

    GError *error = NULL;
    GstElement *bin = gst_parse_bin_from_description(str, TRUE, &error);
    if (error) {
        // Set err_msg
        gchar *tmp = g_strdup_printf("%s. (%s)", error->message, str);
        *err_msg = g_strdup_printf(_("Cannot create audio pipeline. %s.\n"), tmp);
        g_free(tmp);
        g_error_free(error);
        g_free(str);
        goto LBL_1;
    }

    g_free(str);

    GstElement *resample = create_element("audioresample", NULL);
    GstElement *convert = create_element("audioconvert", NULL);

    // Filesink. Caller must set its "location" property.
    GstElement *filesink = create_element("filesink", "filesink");

    gst_bin_add_many(GST_BIN(pipeline), adder, level, resample, convert, bin, filesink, NULL);

    // Link
    if (!gst_element_link_many(adder, level, resample, convert, bin, filesink, NULL)) {
        *err_msg = g_strdup_printf(_("Cannot create audio pipeline. %s.\n"), "Cannot link.");
        goto LBL_1;
    }

    // Now create audio source for all devices
    GList *item = g_list_first(parms->dev_list);
    while (item) {
        // Device name
        const gchar *device = (gchar*)item->data;

        // Source
        const gchar *source_name = (parms->source ? parms->source : "pulsesrc");
        GstElement *source = create_element(source_name, NULL);
        g_object_set(G_OBJECT(source), "device", device, NULL);

        //Queue
        GstElement *queue = create_element("queue", NULL);
        gst_bin_add_many(GST_BIN(pipeline), source, queue, NULL);

        // Link source# -> queue#
        gst_element_link(source, queue);

        // Link queue# -> adder
        // Gstreamer 1.0:
        gst_element_link_pads(queue, NULL, adder, NULL);

        // Next source
        item = g_list_next(item);    
    }

    // Ok
    return pipeline;

LBL_1:
    // Got an error
    return NULL;
}

GstElement *pipeline_create_VAD(PipelineParms *parms, gchar **err_msg) {
    if (!parms) return NULL;
    // Create a GStreamer pipeline for VAD, Voice Activity Detection.

    GstElement *pipeline = NULL;

    // Wash the device list. User may have disconnected microphones and webcams.
    // Invalid devices will crash the GStreamer pipeline. 
    // Remove invalid devices.
    GList *new_list = audio_sources_wash_device_list(parms->dev_list);

    if (g_list_length(new_list) < 1) {
        new_list = g_list_append(new_list, NULL);
    }

    pipeline = pipeline_create_complex_VAD(parms, err_msg);

    // Free new_list
    str_list_free(new_list);
    new_list = NULL;

    return pipeline;
}

#if 0 
static GstElement *pipeline_create_simple_VAD(PipelineParms *parms, gchar **err_msg) {
    // Creater a simple pipelinewwaY TO MAKE for VAD, Voice Activity Detection.
    //
    // Typical pipelines:
    // gst-launch pulsesrc device=alsa_input.usb-Creative_Technology_Cam_Socialize !
    //            cutter name=cutter threshold=0.3 run-length=4000000000 ! fakesink

    // Or using the GstVader element (see .../gst-plugin/src/)
    // gst-launch pulsesrc device=alsa_input.usb-Creative_Technology_Cam_Socialize !
    //            vader name=cutter threshold=0.3 run-length=4000000000 ! fakesink"

    GstElement *pipeline = gst_pipeline_new("Voice Activity Detector");

    // Source
    const gchar *source_name = (parms->source ? parms->source : "pulsesrc");
    GstElement *source = create_element(source_name, NULL);

    // Set device
    const gchar *device = g_list_nth_data(parms->dev_list, 0);
    if (device) {
        g_object_set(G_OBJECT(source), "device", device, NULL);
    }

    // Do we find a GstVader element? (we compile it from source. see .../audio-recorder/gst-plugin/ folder)
    // Vader is a VAD (Voice Activity Detector), it's an improved version of GstCutter.
    // Ref: http://sourceforge.net/projects/cmusphinx/
    // Ref: http://developer.gnome.org/gst-plugins-libs/0.11/gst-plugins-good-plugins-cutter.html
    GstElement *cutter = gst_element_factory_make("vader", "cutter");

    if (!GST_IS_ELEMENT(cutter)) {
        // Standard GstCutter element should do fine
        cutter = create_element("cutter", "cutter");
    }

    // Disable "cutter" messages. Set threshold to 1.0.
    // We set this value later.
    g_object_set(G_OBJECT(cutter), "threshold", 1.0, NULL);

    // Fakesink
    GstElement *fakesink = create_element("fakesink", NULL);

    gst_bin_add_many(GST_BIN(pipeline), source, cutter, fakesink, NULL);

    // Link
    if (!gst_element_link_many(source, cutter, fakesink, NULL)) {
        *err_msg = g_strdup_printf(_("Cannot create audio pipeline. %s.\n"), "Cannot link.");
        goto LBL_1;
    }

    // Ok
    return pipeline;

LBL_1:
    // Got an error
    return NULL;
}
#endif

static GstElement *pipeline_create_complex_VAD(PipelineParms *parms, gchar **err_msg) {
    // Create a complex pipeline for VAD. Use the GstAdder element.
    // Ref: http://gstreamer.freedesktop.org/data/doc/gstreamer/head/gst-plugins-base-plugins/html/gst-plugins-base-plugins-adder.html
    // For 2 more devices.

    // Typical pipelines:
    // $ gst-launch-0.10 adder name=mix
    //        ! level name=level
    //        ! fakesink
    //        { pulsesrc device=alsa_output.pci-0000_04_02.0.analog-stereo.monitor ! queue ! mix. }
    //        { pulsesrc device=alsa_input.pci_8086_24c5_alsa_WebCam ! queue ! mix.
    //
    // $ gst-launch-0.10 adder name=mix ! level ! fakesink { pulsesrc ! queue ! mix. }

    GstElement *pipeline = gst_pipeline_new("Voice Activity Detector");

    // Mix audio from several sources
    GstElement *adder = create_element("adder", "mixer");

    // Level (dB) data
    GstElement *level = create_element("level", "level");

    // Fakesink
    GstElement *fakesink = create_element("fakesink", "fakesink");

    gst_bin_add_many(GST_BIN(pipeline), adder, level, fakesink, NULL);

    // Link
    if (!gst_element_link_many(adder, level, fakesink, NULL)) {
        *err_msg = g_strdup_printf(_("Cannot create audio pipeline. %s.\n"), "Cannot link.");
        goto LBL_1;
    }

    // Now create audio source for all devices
    GList *item = g_list_first(parms->dev_list);
    while (item) {
        // Device name
        const gchar *device = (gchar*)item->data;

        // Source
        const gchar *source_name = (parms->source ? parms->source : "pulsesrc");
        GstElement *source = create_element(source_name, NULL);

        if (device) {
            g_object_set(G_OBJECT(source), "device", device, NULL);
        }

        // Queue
        GstElement *queue = create_element("queue", NULL);

        gst_bin_add_many(GST_BIN(pipeline), source, queue, NULL);

        // Link source# -> queue#
        gst_element_link(source, queue);

        // Link queue# -> adder
        // Gstreamer 1.0:
        gst_element_link_pads(queue, NULL, adder, NULL);

        // Next source
        item = g_list_next(item);    
    }

    // Ok
    return pipeline;

LBL_1:
    // Got an error
    return NULL;
}

#if 0 
gchar *pipeline_get_debug_str(GstElement *pipeline) {
    // Read pipeline and make a string represtation.
    GstIterator *it = gst_bin_iterate_elements(GST_BIN(pipeline));

    gpointer item = NULL;
    GstIteratorResult res = gst_iterator_next(it, &item);
    while (res == GST_ITERATOR_OK) {
        
        gchar *elem_n = gst_element_get_name(item);

        gchar *name = NULL;
        g_object_get(item, "name", &name, NULL);
        g_print("Elem %s, name=%s\n", elem_n, name);
        g_free(name);

        g_free(elem_n);

        res = gst_iterator_next(it, &item);
    }

    gst_iterator_free(it);

    return "nnn";
}

void element_properties (GstElement * element) {
  GParamSpec **property_specs;
  guint num_properties, i;
  gboolean readable;

  g_return_if_fail (element != NULL);

  property_specs = g_object_class_list_properties(G_OBJECT_GET_CLASS (element), &num_properties);

  for (i = 0; i < num_properties; i++) {
    GValue value = { 0, };wwaY TO MAKE
    GParamSpec *param = property_specs[i];

    readable = FALSE;

    g_value_init(&value, param->value_type);

    if (param->flags & G_PARAM_READABLE) {
      g_object_get_property (G_OBJECT (element), param->name, &value);
      readable = TRUE;
    }

    if (!readable) continue;

    switch (G_VALUE_TYPE (&value)) {

      case G_TYPE_STRING:      // String
      {
        GParamSpecString *pstring = G_PARAM_SPEC_STRING (param);
        if (readable) {       
          const char *string_val = g_value_get_string (&value);
          gchar work_string[100];

          if (string_val == NULL)
            sprintf (work_string, "\"%s\"", "null");
          else
            sprintf (work_string, "\"%s\"", string_val);

        } else {
          g_print ("%-*s", c3w, "<not readable>");     
        }
        g_print (" | %-*s", c4w, "G_TYPE_STRING");     

        if (pstring->default_value == NULL)
          g_print (" | %s", "null");
        else
          g_print (" | \"%s\"", pstring->default_value);
        break;
      }

      case G_TYPE_BOOLEAN:     //  Boolean
      {
        GParamSpecBoolean *pboolean = G_PARAM_SPEC_BOOLEAN (param);
        if (readable)           
          g_print ("%-*s", c3w,
              (g_value_get_boolean (&value) ? "true" : "false"));
        else
          g_print ("%-*s", c3w, "<not readable>");
        g_print (" | %-*s", c4w, "G_TYPE_BOOLEAN");     
        g_print (" | %s ",     
            (pboolean->default_value ? "true" : "false"));
        break;
      }

      case G_TYPE_ULONG:       //  Unsigned Long
      {
        GParamSpecULong *pulong = G_PARAM_SPEC_ULONG (param);
        if (readable)          
          g_print ("%-*lu", c3w, g_value_get_ulong (&value));
        else
          g_print ("%-*s", c3w, "<not readable>");
        g_print (" | %-*s", c4w, "G_TYPE_ULONG");       
        g_print (" | (%lu - %lu)   %lu ", pulong->minimum, pulong->maximum,   
            pulong->default_value);    
        break;
      }

      case G_TYPE_LONG:        //  Long
      {
        GParamSpecLong *plong = G_PARAM_SPEC_LONG (param);
        if (readable)           
          g_print ("%-*ld", c3w, g_value_get_long (&value));
        else
          g_print ("%-*s", c3w, "<not readable>");
        g_print (" | %-*s", c4w, "G_TYPE_LONG");
        g_print (" | (%ld - %ld)   %ld ", plong->minimum, plong->maximum,
            plong->default_value);     
        break;
      }

      case G_TYPE_UINT:        //  Unsigned Integer
      {
        GParamSpecUInt *puint = G_PARAM_SPEC_UINT (param);
        if (readable)       
          g_print ("%-*u", c3w, g_value_get_uint (&value));
        else
          g_print ("%-*s", c3w, "<not readable>");
        g_print (" | %-*s", c4w, "G_TYPE_UINT"); 
        g_print (" | (%u - %u)   %u ", puint->minimum, puint->maximum, 
            puint->default_value);     
        break;
      }

      case G_TYPE_INT:         //  Integer
      {
        GParamSpecInt *pint = G_PARAM_SPEC_INT (param);
        if (readable)          
          g_print ("%-*d", c3w, g_value_get_int (&value));
        else
          g_print ("%-*s", c3w, "<not readable>");
        g_print (" | %-*s", c4w, "G_TYPE_INT"); 
        g_print (" | (%d - %d)   %d ", pint->minimum, pint->maximum, 
            pint->default_value);      
        break;
      }

      case G_TYPE_UINT64:      //  Unsigned Integer64.
      {
        GParamSpecUInt64 *puint64 = G_PARAM_SPEC_UINT64 (param);
        if (readable)           
          g_print ("%-*" G_GUINT64_FORMAT, c3w, g_value_get_uint64 (&value));
        else
          g_print ("%-*s", c3w, "<not readable>");
        g_print (" | %-*s", c4w, "G_TYPE_UINT64");
        g_print (" | (%" G_GUINT64_FORMAT " - %" G_GUINT64_FORMAT ")" "   %" G_GUINT64_FORMAT " ", puint64->minimum, puint64->maximum,  
            puint64->default_value); 
        break;
      }

      case G_TYPE_INT64:       // Integer64
      {
        GParamSpecInt64 *pint64 = G_PARAM_SPEC_INT64 (param);
        if (readable) 
          g_print ("%-*" G_GINT64_FORMAT, c3w, g_value_get_int64 (&value));
        else
          g_print ("%-*s", c3w, "<not readable>");
        g_print (" | %-*s", c4w, "G_TYPE_INT64"); 
        g_print (" | (%" G_GINT64_FORMAT " - %" G_GINT64_FORMAT ")" "   %" G_GINT64_FORMAT " ", pint64->minimum, pint64->maximum,
            pint64->default_value); 
        break;
      }

      case G_TYPE_FLOAT:       //  Float.
      {
        GParamSpecFloat *pfloat = G_PARAM_SPEC_FLOAT (param);
        if (readable)
          g_print ("%-*g", c3w, g_value_get_float (&value));
        else
          g_print ("%-*s", c3w, "<not readable>");
        g_print (" | %-*s", c4w, "G_TYPE_FLOAT"); 
        g_print (" | (%g - %g)   %g ", pfloat->minimum, pfloat->maximum, 
            pfloat->default_value);   
        break;
      }

      case G_TYPE_DOUBLE:      //  Double
      {
        GParamSpecDouble *pdouble = G_PARAM_SPEC_DOUBLE (param);
        if (readable)        
          g_print ("%-*g", c3w, g_value_get_double (&value));
        else
          g_print ("%-*s", c3w, "<not readable>");
        g_print (" | %-*s", c4w, "G_TYPE_DOUBLE");  
        g_print (" | (%g - %g)   %g ", pdouble->minimum, pdouble->maximum,  
            pdouble->default_value);  
        break;
      }

      default:
        if (param->value_type == GST_TYPE_CAPS) {
          const GstCaps *caps = gst_value_get_caps (&value);
          if (!caps)
            g_print ("%-*s | %-*.*s |", c3w, "Caps (NULL)", c4w, c4w, " ");
          else {
            gchar prefix_string[100];
            sprintf (prefix_string, "    | %-*.*s | ", c2w, c2w, " ");
            print_caps (caps, prefix_string);
          }
        }


        break;
    }

    if (!readable)
      g_print (" Write only\n");
    else
      g_print ("\n");

    g_value_reset (&value);
  }

  if (0 == num_properties)
    g_print ("  none\n");

  g_free (property_specs);
}

static GstElement *pipeline_create_test(PipelineParms *parms, gchar **err_msg) {
    GstElement *pipeline = NULL;
    gchar *pipeline_cmd = g_strdup_printf ("%s name=source"
                                           " ! level name=level"
                                           " ! queue"
                                           " ! audioconvert"
                                           " ! %s"
                                           " ! filesink name=filesink",
                                           parms->source,
                                           parms->profile_str);

    LOG_DEBUG("\n\n");
    LOG_DEBUG("Going to create pipeline:\ngst-launch %s\n\n", pipeline_cmd);

    GError *err = NULL;
    pipeline = gst_parse_launch(pipeline_cmd, &err);
    g_free(pipeline_cmd);

    if (err) {
        // Set err_msg
        *err_msg = g_strdup_printf(_("Cannot create audio pipeline. %s.\n"), err->message);
        g_error_free(err);
        goto LBL_1;
    }

    // Get the "source0" element
    GstElement *source = gst_bin_get_by_name(GST_BIN(pipeline), "source");

    if (!GST_IS_ELEMENT(source)) {
        // Set err_msg
        *err_msg = g_strdup_printf(_("Cannot find audio element %s.\n"), parms->source);
        goto LBL_1;
    }

    // Set audio device
    if (g_list_length(parms->dev_list) > 0) {
        gchar *device = parms->dev_list->data;
        g_object_set(G_OBJECT(source), "device", device, NULL);
    }

    g_object_unref(source);

    // Ok
    return pipeline;

LBL_1:
    // Got an error
    return NULL;
}
#endif

#if 0 
static GstElement *pipeline_create_TEST(PipelineParms *parms, gchar **err_msg) {
    GstElement *pipeline = NULL;

    gchar *pipeline_cmd = g_strdup_printf ("pulsesrc device=alsa_output.pci-0000_01_00.1.hdmi-stereo.monitor"
                                     " ! queue"
                                     " ! level name=level"
                                     " ! audioconvert"
                                     " ! audio/x-raw,rate=44100,channels=2 ! vorbisenc name=enc quality=0.5 ! oggmux"
                                     " ! filesink name=filesink");

    LOG_DEBUG("\n\n");
    LOG_DEBUG("Going to create pipeline:\ngst-launch %s\n\n", pipeline_cmd);

    GError *err = NULL;
    pipeline = gst_parse_launch(pipeline_cmd, &err);
    g_free(pipeline_cmd);

    if (err) {
        // Set err_msg
        *err_msg = g_strdup_printf(_("Cannot create audio pipeline. %s.\n"), err->message);
        g_error_free(err);
        goto LBL_1;
    }

    // Ok
    return pipeline;

LBL_1:
    // Got an error
    return NULL;
}
#endif

