/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License (GPL3), or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Library General Public License 3 for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License 3 along with this program; if not, see /usr/share/common-licenses/GPL file
 * or write to Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/
#include "gtklevelbar.h"
#include <math.h> // round()

// A simple level bar widget.
// By Osmo Antero.
//
// Sample call:
// GtkWidget *lb = gtk_level_bar_new();
// gtk_widget_show(lb);
//
// Set value [0.0, 1.0].
// gtk_level_bar_set_fraction(GTK_LEVEL_BAR(lb), 0.8);
//
// Show %-value [0 - 100] or plain value [0 - 1.0] on the levelbar (one of: SCALE_NONE, SCALE_VALUE, SCALE_PRECENT)
// gtk_level_bar_set_scale(GTK_LEVEL_BAR(lb), SCALE_PRECENT);
//

struct _GtkLevelBarPrivate {
    gdouble fraction;
    guint bar_height;
    enum SCALE_TYPE scale_type;
};

static void gtk_level_bar_get_preferred_width(GtkWidget *widget, gint *minimum, gint *natural);
static void gtk_level_bar_get_preferred_height (GtkWidget *widget,gint *minimum, gint *natural);

static void gtk_level_bar_real_update(GtkLevelBar *progress);
static gboolean gtk_level_bar_draw(GtkWidget *widget, cairo_t *cr);

static void gtk_level_bar_finalize(GObject *object);

G_DEFINE_TYPE(GtkLevelBar, gtk_level_bar, GTK_TYPE_WIDGET);

static void gtk_level_bar_class_init (GtkLevelBarClass *class) {
    GObjectClass *gobject_class;
    GtkWidgetClass *widget_class;

    gobject_class = G_OBJECT_CLASS (class);
    widget_class = (GtkWidgetClass *)class;

    gobject_class->set_property = NULL;
    gobject_class->get_property = NULL;
    gobject_class->finalize = gtk_level_bar_finalize;

    widget_class->draw = gtk_level_bar_draw;
    widget_class->get_preferred_width = gtk_level_bar_get_preferred_width;
    widget_class->get_preferred_height = gtk_level_bar_get_preferred_height;

    g_type_class_add_private (class, sizeof (GtkLevelBarPrivate));
}

static void gtk_level_bar_init(GtkLevelBar *pbar) {
    GtkLevelBarPrivate *priv;

    pbar->priv = G_TYPE_INSTANCE_GET_PRIVATE(pbar, GTK_TYPE_LEVEL_BAR, GtkLevelBarPrivate);
    priv = pbar->priv;

    priv->fraction = 0.0;
    priv->bar_height = 8;
    priv->scale_type = SCALE_VALUE;

    gtk_widget_set_has_window(GTK_WIDGET (pbar), FALSE);
}

GtkWidget *gtk_level_bar_new(void) {
    GtkWidget *pbar;
    pbar = g_object_new(GTK_TYPE_LEVEL_BAR, NULL);
    return pbar;
}

static void gtk_level_bar_real_update (GtkLevelBar *pbar) {
    GtkWidget *widget;

    g_return_if_fail (GTK_IS_LEVEL_BAR (pbar));

    GtkLevelBarPrivate __attribute__ ((unused)) *priv = pbar->priv;

    widget = GTK_WIDGET(pbar);

    gtk_widget_queue_draw(widget);
}

static void gtk_level_bar_finalize (GObject *object) {
    G_OBJECT_CLASS(gtk_level_bar_parent_class)->finalize (object);
}

static void gtk_level_bar_get_preferred_width (GtkWidget *widget,gint *minimum, gint *natural) {
    *minimum = 50;
    *natural = 160;
}

static void gtk_level_bar_get_preferred_height (GtkWidget *widget, gint      *minimum, gint      *natural) {
    *minimum = 6;
    *natural = 8;
}

static gboolean gtk_level_bar_draw(GtkWidget *widget, cairo_t *cr) {
    // Draw level bar and optional text
    GtkLevelBar *pbar = GTK_LEVEL_BAR (widget);
    GtkLevelBarPrivate *priv = pbar->priv;
    GtkStyleContext *context;
    int width, height;

    context = gtk_widget_get_style_context(widget);

    width = gtk_widget_get_allocated_width(widget);
    height = gtk_widget_get_allocated_height(widget);

    // Bar thickness
    gdouble bar_height = MIN(height , priv->bar_height);

    // Vertical pos
    gdouble y = (height - bar_height)/2;

    // Pulse width
    gdouble w = priv->fraction/(1.00/width);

    gtk_style_context_save(context);
    // gtk_style_context_add_class(context, GTK_STYLE_CLASS_TROUGH);
    gtk_render_background(context, cr, 0, 0, width, height);
    gtk_render_frame(context, cr, 0, 0, width, height);

    // Progressbar style
    gtk_style_context_add_class(context, GTK_STYLE_CLASS_PROGRESSBAR);

    if (priv->fraction > 0.001) {
        // Render activity bar with current theme and colors
        gtk_render_activity(context, cr, 0, y, w, bar_height);
    }

    gtk_style_context_restore(context);

    cairo_select_font_face(cr, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
    cairo_set_font_size(cr, priv->bar_height);

    GdkRGBA color;
    gtk_style_context_get_border_color(context, GTK_STATE_NORMAL, &color);
    gdk_cairo_set_source_rgba(cr, &color);
    color.alpha = 0.9;

    // Calculate total width of scale
    cairo_text_extents_t extents;
    cairo_text_extents(cr, "0.0", &extents);
    gint total_w = 9 * (extents.x_advance + extents.width);

    // Debug:
    // g_print("Bar width=%d  total_w=%d  bearing=%3.1f advance=%3.1f char.width=%3.1f\n", width, total_w, 
    //          extents.x_bearing, extents.x_advance, extents.width);

    // Draw all or only each second value?
    gboolean draw_all = (total_w - extents.width) < width; 

    // Draw normalized values [0 - 1.0]?
    if (priv->scale_type == SCALE_VALUE) {
        // Scale: 0.1  0.2  0.3  0.4...0.9
        gint i = 0;
        for (i=0; i < 10; i++) {
            gchar *s = NULL;

            if (draw_all || (i % 2 == 0))
                s = g_strdup_printf("%2.1f", (gdouble)i/10.0);

            if (!s) continue;

            cairo_text_extents_t extents;
            cairo_text_extents(cr, s, &extents);

            gdouble xx = (width/10) * i;
            gdouble yy = height/2-(extents.height/2 + extents.y_bearing);

            cairo_move_to(cr, xx, yy);
            cairo_show_text(cr, s);

            g_free(s);        
        }

    // Draw percentage scale?
    } else if (priv->scale_type == SCALE_PERCENT) {
        // Scale: 10% . 20% . 30% . 40% . 50%...90%
        gint i = 0;
        for (i=0; i < 10; i++) {
            gchar *s = NULL;
            if (i % 2 == 0)
                s = g_strdup_printf("%2.0f%%", (gdouble)i*10.0);
            else
                s = g_strdup_printf("%3s", ".");

            cairo_text_extents_t extents;
            cairo_text_extents(cr, s, &extents);

            gdouble xx = (width/10) * i;
            gdouble yy = height/2-(extents.height/2 + extents.y_bearing);

            cairo_move_to(cr, xx, yy);
            cairo_show_text(cr, s);

            g_free(s);
        }
    }

    #if 0 
    // Commented out by moma 30.sep.2012.

    // Set text
    if (priv->text) {
        cairo_select_font_face(cr, "Sans", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);

        cairo_set_font_size(cr, 0.5*height);
        cairo_text_extents_t extents;
        cairo_text_extents(cr, priv->text, &extents);

        // Ref: http://cairographics.org/manual/cairo-cairo-scaled-font-t.html#cairo-text-extents-t
        gdouble xx = width-(extents.width + extents.x_bearing)-2;
        gdouble yy = height/2-(extents.height/2 + extents.y_bearing);

        GdkRGBA color;
        gtk_style_context_get_border_color(context, GTK_STATE_NORMAL, &color);
        gdk_cairo_set_source_rgba(cr, &color);
        color.alpha = 0.9;

        cairo_move_to(cr, xx, yy);
        cairo_show_text(cr, priv->text);
    }
    #endif

    return FALSE;
}

void gtk_level_bar_set_fraction(GtkLevelBar *pbar, gdouble fraction) {
    // Set fraction [0.0, 1.0]
    GtkLevelBarPrivate* priv;
    g_return_if_fail (GTK_IS_LEVEL_BAR (pbar));
    priv = pbar->priv;

    priv->fraction = CLAMP(fraction, 0.0, 1.0);
    gtk_level_bar_real_update (pbar);
}

gdouble gtk_level_bar_get_fraction(GtkLevelBar *pbar) {
    // Get fraction
    g_return_val_if_fail(GTK_IS_LEVEL_BAR (pbar), 0);
    return pbar->priv->fraction;
}

void gtk_level_bar_set_bar_height(GtkLevelBar *pbar, guint height) {
    // Set bar height (thickness). Normally 8 pixels.
    g_return_if_fail(GTK_IS_LEVEL_BAR (pbar));
    GtkLevelBarPrivate* priv = pbar->priv;
    priv->bar_height = height;
    // Redraw
    gtk_level_bar_real_update(pbar);
}

guint gtk_level_bar_get_bar_height(GtkLevelBar *pbar) {
    // Get bar thickness
    g_return_val_if_fail(GTK_IS_LEVEL_BAR(pbar), 0);
    return pbar->priv->bar_height;
}

void gtk_level_bar_set_scale(GtkLevelBar *pbar, enum SCALE_TYPE scale_type) {
    // Set SCALE_TYPE
    g_return_if_fail(GTK_IS_LEVEL_BAR(pbar));
    GtkLevelBarPrivate* priv = pbar->priv;
    priv->scale_type = scale_type;
    // Redraw
    gtk_level_bar_real_update(pbar);
}

enum SCALE_TYPE gtk_level_bar_get_scale(GtkLevelBar *pbar) {
    // Get SCALE_TYPE
    g_return_val_if_fail(GTK_IS_LEVEL_BAR(pbar), SCALE_NONE);
    GtkLevelBarPrivate* priv = pbar->priv;
    return priv->scale_type;
}

