/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License (GPL3), or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Library General Public License 3 for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License 3 along with this program; if not, see /usr/share/common-licenses/GPL file
 * or write to Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/
#include "media-profiles.h"
#include "log.h"

// EDIT 12.feb.2013: 
// Removed dependency to libgnome-media-profiles-dev and libgnome-media-profiles-3.0-0 libraries.
// Seems like these libraries will be replaced by something else.
// Please see: https://bugs.launchpad.net/audio-recorder/+bug/1121376

// We are using hard coded values for media types.
 ProfileRec DefaultProfiles[] = {
 {"aac",       "CD Quality, AAC",     "m4a", "audio/x-raw,rate=44100,channels=2 ! faac profile=2 ! ffmux_mp4"},
 {"cdlossless","CD Quality, Lossless","flac","audio/x-raw,rate=44100,channels=2 ! flacenc name=enc"},
 {"cdlossy",   "CD Quality, Lossy",   "ogg", "audio/x-raw,rate=44100,channels=2 ! vorbisenc name=enc quality=0.5 ! oggmux"},
 {"mp2",       "CD Quality, MP2",     "mp2", "audio/x-raw,rate=44100,channels=2 ! twolame name=enc mode=0 bitrate=192 ! id3v2mux"},
 {"mp3",       "CD Quality, MP3",     "mp3", "audio/x-raw,rate=44100,channels=2 ! lamemp3enc name=enc target=0 quality=6 ! xingmux ! id3v2mux"},
 {"voicelossless","Voice, Lossless",  "wav", "audio/x-raw,rate=22050,channels=1 ! wavenc name=enc"},
 {"voicelossy",   "Voice, Lossy",     "spx", "audio/x-raw,rate=32000,channels=1 ! speexenc name=enc ! oggmux"},
};

static gboolean profiles_test_pipeline(const gchar *pipe);

static GList *profiles_get_active();

// List of active media profiles 
static GList *g_profile_list = NULL;

void media_profiles_init() {
    LOG_DEBUG("Init media-profiles.c.\n");
}

void media_profiles_exit() {
    LOG_DEBUG("Clean up media-profiles.c.\n");

    // Clean up. Free the list, do not touch data items.
    if (g_profile_list) {
        g_list_free(g_profile_list);
    }
    g_profile_list = NULL;
}

static void profiles_load() {
    // Load active media profiles (audio formats)
    if (g_profile_list) return;

    g_profile_list = profiles_get_active();
}

static gboolean profiles_test_pipeline(const gchar *pipe) {
    // Test if system has support for given audio format (GStreamer-plugin has been installed). 
    // Create and test a basic pipeline.

    // Pipeline to test
    gchar *test = g_strdup_printf ("fakesrc ! %s ! fakesink", pipe);

    gboolean pipe_OK = FALSE;

    // Parse pipeline
    GError *error = NULL;
    GstElement *p = gst_parse_launch(test, &error);
    if (error) {
        g_error_free(error);
    } else {
        pipe_OK = TRUE;
    }

    if (GST_IS_OBJECT(p)) {
        gst_object_unref(p);
    }

    g_free(test);

    return pipe_OK;
} 

static GList *profiles_get_active() {
    // Return a list of active audio profiles.
    GList *list = NULL;

    guint i = 0;
    for (i = 0; i < sizeof(DefaultProfiles) / sizeof(DefaultProfiles[0]); i++) {

        // Test pipeline
        const gchar *pipe = DefaultProfiles[i].pipe;

        gboolean pipe_OK = profiles_test_pipeline(pipe);

        if (pipe_OK) {
            // It's OK. Add record to the list. 
            list = g_list_append(list, &DefaultProfiles[i]);
        }
    }

    return list;
}

const ProfileRec *profiles_find_profile(const gchar *id) {

    profiles_load();

    GList *item = g_list_first(g_profile_list);
    while (item) {
        ProfileRec *p = (ProfileRec*)item->data;
        if (!g_strcmp0(p->id, id)) {
            return p;
        } 

        item = g_list_next(item);
    }
    return NULL;
}

gchar *profiles_get_extension(const gchar *id) {
    // Return file extension (such as .mp3, .ogg) for the given profile

    const ProfileRec *p = profiles_find_profile(id);
    if (!p) return NULL;

    // Caller should g_free() this value
    return g_strdup(p->ext);
}

gchar *profiles_get_pipeline(const gchar *id) {
    // Return pipeline fragment for the given profile

    const ProfileRec *p = profiles_find_profile(id);
    if (!p) return NULL;

    // Caller should g_free() this value
    return g_strdup(p->pipe);
}

gboolean profiles_check_profile(const gchar *id) {
    // Check if the id is a valid
    const ProfileRec *p = profiles_find_profile(id);
    return (p != NULL);
}

static void profiles_add_data(GtkListStore *store) {
    // Fill GtkListStore with active audio formats; ids and names.

    // Load activate profiles
    profiles_load();

    GList *item = g_list_first(g_profile_list);
    while (item) {
        ProfileRec *p = (ProfileRec*)item->data;

        // Typical audio format list is
        //  CD Quality, AAC (.m4a)
        //  CD Quality, Lossless (.flac)
        //  CD Quality, Lossy (.ogg)
        //  CD Quality, MP2 (.mp2)
        //  CD Quality, MP3 (.mp3)
        //  Voice, Lossless (.wav)
        //  Voice, Lossy (.spx)

        gchar *e = g_utf8_strup(p->ext, -1);
        gchar *n = g_utf8_strdown(p->name, -1);
        gchar *txt = g_strdup_printf(".%s  (%s)", e, n);

        GtkTreeIter iter;

        // Add OGG type "CD Quality, Lossy" as first (becomes default)
        if (!g_strcmp0(p->ext, "ogg") || !g_strcmp0(p->ext, "oga")) {
            gtk_list_store_prepend(GTK_LIST_STORE(store), &iter);
        } else {
            // Add last
            gtk_list_store_append(GTK_LIST_STORE(store), &iter);
        }

        gtk_list_store_set(GTK_LIST_STORE(store), &iter, 0, p->id, 1, txt, -1);

        g_free(txt);
        g_free(e);
        g_free(n);

        item = g_list_next(item);
    }
}

gchar *profiles_get_selected_id(GtkWidget *widget) {
    // Return selected profile id

    GtkComboBox *combo = GTK_COMBO_BOX(widget);
    g_return_val_if_fail(GTK_IS_COMBO_BOX(combo), NULL);

    GtkTreeIter iter;
    gchar *id = NULL;

    if (!gtk_combo_box_get_active_iter(combo, &iter)) return NULL;

    gtk_tree_model_get(gtk_combo_box_get_model(combo), &iter, COL_PROFILE_ID, &id, -1);

    // The caller should g_free() this value
    return id;
}

GtkWidget *profiles_create_combobox() {
    // Create a GtkComboBox with N_PROFILE_COLUMNS. Populate it with supported audio formats.

    // Create store
    GtkListStore *store = gtk_list_store_new(N_PROFILE_COLUMNS, G_TYPE_STRING, G_TYPE_STRING);

    // Add data to the listbox
    profiles_add_data(store);

    GtkWidget *combo = gtk_combo_box_new_with_model(GTK_TREE_MODEL(store));

    // Unref store
    g_object_unref(store);

    // "Id" column, invisible
    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
    g_object_set(cell, "visible", 0, NULL);
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), cell, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo), cell, "text", COL_PROFILE_ID, NULL);

    // "Name" column, visible
    cell = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), cell, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo), cell, "text", COL_PROFILE_TXT, NULL);

    return combo;
}

